#include "jserver.h"

thread session_create(jid owner, session_onSend evt, void *evt_data)
{
    /* creats a new session, registers the resource for it, and spawns a thread to handle it */
    thread t;
    session s;
    xmlnode presence;
    jid user;
    udata u;

    if(owner == NULL || owner->resource == NULL) /* illegal */
        return NULL;
 
    debug("session_create");
 
    /* create session thread */
    t = thread_new();
    s = pmalloc(t->p, sizeof(struct session_struct));
    t->data = (void *)s;

    /* setup internal communication points */
    s->mp_from = pth_msgport_create("session_from");
    s->mp_to = pth_msgport_create("session_to");
    s->send_evt = evt;
    s->send_evt_data = evt_data;

    /* session identity and default settings */
    s->id = jid_new(t->p, jid_full(owner));
    s->r = resource_create(RESOURCE_SESSION, owner->user, owner->resource, s->mp_to);
    s->p = t->p;
    s->exit_flag = 0;
    s->presence = xmlnode_new_tag("presence");
    xmlnode_put_attrib(s->presence,"type","unavailable");

    /* only for non-anonymous users */
    if(owner->user != NULL)
    {
        s->u = user_data(owner->user);
        if(s->u->primary != NULL)
        {
            /* we'll be default since we're the most recent */
            if(s->u->priority == 0)
            {
                s->u->primary = s->r;
                s->u->priority = 0;
            }

            /* built-in presence broadcast to any of this users other sessions only */
            presence = xmlnode_new_tag("presence");
            user = jid_new(t->p,jid_full(owner));
            user->resource = NULL; /* broadcast */
            xmlnode_put_attrib(presence,"to",jid_full(user));
            xmlnode_put_attrib(presence,"from",jid_full(owner));
            xmlnode_put_attrib(presence,"type","available");
            deliver(jpacket_new(presence));
        }else{
            s->u->primary = s->r;
            s->u->priority = 0;
        }
    }else{
        s->u = NULL;
    }

    /* start thread */
    t->id = pth_spawn(PTH_ATTR_DEFAULT, session_main, (void *)t);
    pth_yield(NULL); /* let it have a chance to start up, XXX NULL should be t->id but it seems to be broken in pth */
 
    return t;
}

void session_end(session s)
{
    jpq q;
    xmlnode presence;

    if(s == NULL)
        return;

    /* flag the session to exit ASAP */
    s->exit_flag = 1;

    /* make sure we're not the primary session */
    if(s->u != NULL && s->u->primary == s->r)
    {
        s->u->primary = NULL;
        s->u->priority = 0;
    }

    /* if this session still thinks it's online, tell it otherwise */
    q = pmalloc(s->p, sizeof(_jpq));
    if(j_strcmp(xmlnode_get_attrib(s->presence,"type"),"unavailable") != 0)
    {
        presence = xmlnode_new_tag("presence");
        xmlnode_put_attrib(presence,"type","unavailable");
        xmlnode_insert_cdata(xmlnode_insert_tag(presence,"status"),"disconnected",12);
        q->p = jpacket_new(presence);
    }else{
        /* send it a dummy message so it gets CPU time */
        q->p = NULL;
    }
    pth_msgport_put(s->mp_from, (pth_message_t *)q);
}

void *session_main(void *arg)
{
    thread t = (thread)arg;
    session s = (session)(t->data);
    pth_event_t from, to, ring;
    jpq q;
    xmlnode presence;
    mapi m;
    jid to_id;
    int flag;

    debug("THREAD:SESSION starting");

    from = pth_event(PTH_EVENT_MSG,s->mp_from);
    to = pth_event(PTH_EVENT_MSG,s->mp_to);
    ring = pth_event_concat(from, to, NULL);

    module_dispatch(mapi_new(M_PHASE_SESSION_ON,t->p,s,NULL));

    while(1)
    {
        pth_wait(ring);

        while(1)
        {
            /* this loop looks for messages from our user that need to be delivered elsewhere */
            q = (jpq)pth_msgport_get(s->mp_from);
            if(q == NULL || q->p == NULL)
                break;

            debug("THREAD:SESSION:FROM received a message!");

            if(s->r->user != NULL)
            {
                switch(q->p->type)
                {
                    case JPACKET_MESSAGE:
                        m = mapi_new(M_PHASE_MESSAGE_OUT,q->p->p,s,q->p);
                        module_dispatch(m);
                        if(m->result != M_DEFAULT)
                        {
                            xmlnode_free(q->p->x);
                            continue;
                        }
                        break;
                    case JPACKET_PRESENCE:
                        /* it was a target send */
                        if(q->p->to != NULL)
                            break;

                        /* ha ha, we're better than you, give us default messages! */
                        if(presence_priority(q->p->x) >= s->u->priority)
                        {
                            s->u->priority = presence_priority(q->p->x);
                            s->u->primary = s->r;
                        }else if(s->u->primary == s->r){ /* we were default, but we're lower now :( */
                            s->u->priority = presence_priority(q->p->x);
                        }

                        /* force broadcast to our other resources */
                        presence = xmlnode_dup(q->p->x);
                        xmlnode_put_attrib(presence,"from",jid_full(s->id));
                        to_id = jid_new(q->p->p,jid_full(s->id));
                        to_id->resource = NULL;
                        xmlnode_put_attrib(presence,"to",jid_full(to_id));
                        deliver(jpacket_new(presence));

                        m = mapi_new(M_PHASE_PRESENCE_OUT,q->p->p,s,q->p);
                        module_dispatch(m);
                        if(m->result == M_DEFAULT)
                        { /* module didn't handle it, we'll use our internal roster */
                            to_id = s->u->subscribers;
                            flag = j_strcmp(xmlnode_get_attrib(q->p->x,"type"),"available");
                            if(flag != 0)
                                flag = j_strcmp(xmlnode_get_attrib(q->p->x,"type"),"unavailable");
                            while(to_id != NULL)
                            {
                                if(flag != 0 && presence_check(CMD_GET,jid_full(to_id)) == 0)
                                { /* if this isn't an [un]available type, only send to those we know are available */
                                    to_id = to_id->next;
                                    continue;
                                }
                                to_id->resource = NULL; /* only announce presence to users */
                                presence = xmlnode_dup(q->p->x);
                                xmlnode_put_attrib(presence,"from",jid_full(s->id));
                                xmlnode_put_attrib(presence,"to",jid_full(to_id));
                                deliver(jpacket_new(presence));
                                to_id = to_id->next;
                            }
                        }

                        /* replace this sessions current status */
                        xmlnode_free(s->presence);
                        s->presence = q->p->x;
                        continue;
                    case JPACKET_IQ:
                        /* it was a target send */
                        if(q->p->to != NULL)
                            break;

                        /* known IQ handlers or let module handle it */
                        if(j_strcmp(xmlnode_get_attrib(xmlnode_get_tag(q->p->x,"query"),"xmlns"),"jabber:iq:roster") == 0)
                            session_roster_handle(s,q->p);
                        else
                            module_dispatch(mapi_new(M_PHASE_IQ_HANDLE,q->p->p,s,q->p));

                        /* reply */
                        xmlnode_put_attrib(q->p->x,"type","result");
                        pth_msgport_put(s->mp_to, (pth_message_t *)q);
                        continue;
                }
            }

            /* make sure we have our from set correctly for outgoing crap */
            xmlnode_put_attrib(q->p->x,"from",jid_full(s->id));
            q->p->from = jid_new(q->p->p,jid_full(s->id));

            /* pass these to the general delivery function */
            deliver(q->p);
        }

        while(1)
        {
            /* this loop looks for messages to our user that need to be sent to them */
            q = (jpq)pth_msgport_get(s->mp_to);
            if(q == NULL || q->p == NULL)
                break;

            debug("THREAD:SESSION:TO received data from %s!",jid_full(q->p->from));

            /* XXX if(s->exit_flag) we're exiting and should bounce these suckers! */
            if(s->exit_flag)
            {
                xmlnode_free(q->p->x);
                continue;
            }

            if(s->r->user != NULL)
            {
                switch(q->p->type)
                {
                    case JPACKET_MESSAGE:
                        m = mapi_new(M_PHASE_MESSAGE_IN,q->p->p,s,q->p);
                        module_dispatch(m);
                        if(m->result != M_DEFAULT)
                        {
                            xmlnode_free(q->p->x);
                            continue;
                        }
                        break;
                    case JPACKET_PRESENCE:
                        /* we don't want to talk to ourselves :) */
                        if(jid_cmp(q->p->from,s->id) == 0)
                        {
                            xmlnode_free(q->p->x);
                            continue;
                        }

                        /* one of our sessions is changing presence, internal broadcast */
                        if(j_strcmp(q->p->from->user, s->id->user) == 0 && config_names(CMD_CHECK,q->p->from->server) != NULL)
                        {
                            /* if they lowered, recheck the default in case were better */
                            if(presence_priority(s->presence) > s->u->priority || s->u->primary == NULL)
                            {
                                s->u->priority = presence_priority(s->presence);
                                s->u->primary = s->r;
                            }
                            /* reply */
                            if(j_strcmp(xmlnode_get_attrib(q->p->x,"type"),"available") == 0 || j_strcmp(xmlnode_get_attrib(q->p->x,"type"),"poll") == 0)
                            {
                                presence = xmlnode_dup(s->presence);
                                xmlnode_put_attrib(presence,"to",jid_full(q->p->from));
                                xmlnode_put_attrib(presence,"from",jid_full(s->id));
                                deliver(jpacket_new(presence));                            
                            }
                            break;
                        }

                        m = mapi_new(M_PHASE_PRESENCE_IN,q->p->p,s,q->p);
                        module_dispatch(m);
                        if(m->result == M_DEFAULT)
                        { /* the module didn't handle responding, we will */
                            if(j_strcmp(xmlnode_get_attrib(q->p->x,"type"),"available") == 0 || j_strcmp(xmlnode_get_attrib(q->p->x,"type"),"poll") == 0)
                            {
                                to_id = s->u->subscribers;
                                while(to_id != NULL)
                                {
                                    if(jid_cmp(q->p->from,to_id) == 0)
                                        break;
                                    to_id = to_id->next;
                                }
                                if(to_id != NULL)
                                { /* only reply if they are on our subscription */
                                    presence = xmlnode_dup(s->presence);
                                    xmlnode_put_attrib(presence,"to",jid_full(q->p->from));
                                    xmlnode_put_attrib(presence,"from",jid_full(s->id));
                                    deliver(jpacket_new(presence));
                                }
                                /* polls don't get delivered */
                                if(j_strcmp(xmlnode_get_attrib(q->p->x,"type"),"poll") == 0)
                                    continue;
                            }else if(j_strcmp(xmlnode_get_attrib(q->p->x,"type"),"subscribe") == 0){
                                /* XXX options on the account, autosubscribe? */
                            }
                        }
                        break;
                    case JPACKET_IQ:
                        m = mapi_new(M_PHASE_IQ,q->p->p,s,q->p);
                        module_dispatch(m);
                        if(m->result == M_HANDLED)
                        {
                            /* return the result to the sender */
                            xmlnode_put_attrib(q->p->x,"type","result");
                            xmlnode_put_attrib(q->p->x,"to",jid_full(q->p->from));
                            xmlnode_put_attrib(q->p->x,"from",jid_full(s->id));
                            deliver(jpacket_new(q->p->x));
                        }
                        if(m->result != M_DEFAULT) /* send to user unless we took care of it already */
                        {
                            xmlnode_free(q->p->x);
                            continue;
                        }
                        break;
                }
            }

            /* deliver outgoing for this session to the onSend event passed when this session was created */
            (s->send_evt)(t, q->p);
        }

        if(s->exit_flag)
            break;
    }

    debug("THREAD:SESSION exiting");

    module_dispatch(mapi_new(M_PHASE_SESSION_OFF,t->p,s,NULL));

    /* unlist ourselves */
    resource_free(s->r);

    /* force our other sessions to know we're gone */
    xmlnode_put_attrib(s->presence,"from",jid_full(s->id));
    s->id->resource = NULL; /* broadcast */
    xmlnode_put_attrib(s->presence,"to",jid_full(s->id));
    xmlnode_put_attrib(s->presence,"type","unavailable");
    deliver(jpacket_new(s->presence));

    pth_event_free(ring,PTH_FREE_ALL);
    pth_msgport_destroy(s->mp_from);
    pth_msgport_destroy(s->mp_to);

    pool_free(t->p);
}

